//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright Contributors to the OpenEXR Project.
//

void testNormalizedConversionError();
void testDenormalizedConversionError();
void testRoundingError();
